import asyncio
from logging import Logger
from os.path import join
from re import search

from aiofiles import open as aopen
from asyncpg import create_pool

from pgbelt.config.models import DbupgradeConfig
from pgbelt.util.asyncfuncs import isfile
from pgbelt.util.asyncfuncs import listdir
from pgbelt.util.asyncfuncs import makedirs
from pgbelt.util.postgres import table_empty

RAW = "schema"
NO_INVALID = "no_invalid_constraints"
ONLY_INVALID = "invalid_constraints"


def schema_dir(db: str, dc: str) -> str:
    return f"schemas/{dc}/{db}"


def schema_file(db: str, dc: str, name: str) -> str:
    return join(schema_dir(db, dc), f"{name}.sql")


def table_dir(db: str, dc: str) -> str:
    return f"tables/{dc}/{db}"


def table_file(db: str, dc: str, name: str) -> str:
    return join(table_dir(db, dc), f"{name}.sql")


def _parse_dump_commands(out: str, src_owner: str, dst_owner: str) -> list[str]:
    """
    Given a string containing output from pg_dump, return a list of strings where
    each is a complete postgres command. Commands may be multi-line.
    """
    lines = out.split("\n")
    commands = []

    for line in lines:
        stripped = line.strip()
        # if the line is whitespace only or a comment then ignore it
        if not stripped or stripped.startswith("--"):
            continue

        # replace source owner user with target owner user
        line = line.replace(src_owner, dst_owner)

        # if the last command is terminated or we don't have any yet start a new one
        if not commands or commands[-1].endswith(";\n"):
            commands.append(line + "\n")
        # otherwise we append to the last command because it must be multi-line
        else:
            commands[-1] += line + "\n"

    return commands


async def _execute_subprocess(
    command: list[str], finished_log: str, logger: Logger
) -> bytes:
    p = await asyncio.create_subprocess_exec(
        command[0],
        *command[1:],
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    out, err = await p.communicate()

    if p.returncode != 0:
        raise Exception(
            f"Couldn't do {command}, got code {p.returncode}.\n  out: {out.decode('utf-8')}\n  err: {err.decode('utf-8')}"
        )
    logger.debug(finished_log)
    return out


async def dump_source_tables(
    config: DbupgradeConfig, tables: list[str], logger: Logger
) -> None:
    try:
        await makedirs(table_dir(config.db, config.dc))
    except FileExistsError:
        pass

    logger.info(f"Dumping tables {tables}")

    dumps = []
    for table in tables:
        dumps.append(
            _execute_subprocess(
                [
                    "pg_dump",
                    "--data-only",
                    f"--table={table}",
                    "-Fc",
                    "-f",
                    table_file(config.db, config.dc, table),
                    config.src.pglogical_dsn,
                ],
                f"dumped {table}",
                logger,
            )
        )

    await asyncio.gather(*dumps)


async def load_dumped_tables(
    config: DbupgradeConfig, tables: list[str], logger: Logger
) -> None:
    # unless we get an explicit list of tables to load just load all the dump files
    if not tables:
        tables_dir = table_dir(config.db, config.dc)
        tables = [
            f.split(".")[0]
            for f in await listdir(tables_dir)
            if await isfile(join(tables_dir, f))
        ]

    logger.info(f"Loading dumped tables {tables}")

    # only load a dump file if the target table is completely empty
    async with create_pool(config.dst.root_uri, min_size=1) as pool:
        to_load = []
        for t in tables:
            if await table_empty(pool, t, logger):
                to_load.append(table_file(config.db, config.dc, t))
            else:
                logger.warning(
                    f"Not loading {t}, table not empty. If this is unexpected please investigate."
                )

    loads = []
    for file in to_load:
        loads.append(
            _execute_subprocess(
                [
                    "pg_restore",
                    "-d",
                    config.dst.owner_dsn,
                    file,
                ],
                f"loaded {file}",
                logger,
            )
        )

    await asyncio.gather(*loads)


async def dump_source_schema(config: DbupgradeConfig, logger: Logger) -> None:
    """
    Dump the schema from the source db and write a file with the complete schema,
    one with only the NOT VALID constraints from the schema, and one with everything
    but the NOT VALID constraints.
    """
    logger.info("Dumping schema...")

    command = [
        "pg_dump",
        "--schema-only",
        "--no-owner",
        "-n",
        "public",
        config.src.pglogical_dsn,
    ]

    out = await _execute_subprocess(command, "Retrieved source schema", logger)

    commands_raw = _parse_dump_commands(
        out.decode("utf-8"), config.src.owner_user.name, config.dst.owner_user.name
    )

    commands = []
    for c in commands_raw:
        if "EXTENSION " not in c and "GRANT " not in c and "REVOKE " not in c:
            commands.append(c)

    try:
        await makedirs(schema_dir(config.db, config.dc))
    except FileExistsError:
        pass

    async with aopen(schema_file(config.db, config.dc, RAW), "w") as out:
        for command in commands:
            await out.write(command)

    async with aopen(schema_file(config.db, config.dc, ONLY_INVALID), "w") as out:
        for command in commands:
            if "NOT VALID" in command:
                await out.write(command)

    async with aopen(schema_file(config.db, config.dc, NO_INVALID), "w") as out:
        for command in commands:
            if "NOT VALID" not in command:
                await out.write(command)

    logger.debug("Finished dumping schema.")


async def apply_target_schema(config: DbupgradeConfig, logger: Logger) -> None:
    """
    Load the schema dumped from the source into the target excluding NOT VALID constraints.
    """
    logger.info("Loading schema without constraints...")

    command = [
        "psql",
        config.dst.owner_dsn,
        "-f",
        schema_file(config.db, config.dc, NO_INVALID),
    ]

    await _execute_subprocess(command, "Finished loading schema.", logger)


async def dump_dst_not_valid_constraints(
    config: DbupgradeConfig, logger: Logger
) -> None:
    """
    Dump NOT VALID Constraints from the target database.
    Used when schema is loaded in outside of pgbelt.
    """

    logger.info("Dumping target NOT VALID constraints...")

    command = [
        "pg_dump",
        "--schema-only",
        "--no-owner",
        "-n",
        "public",
        config.dst.pglogical_dsn,
    ]

    out = await _execute_subprocess(command, "Retrieved target schema", logger)

    # No username replacement needs to be done, so replace dst user with the same.
    commands_raw = _parse_dump_commands(
        out.decode("utf-8"), config.dst.owner_user.name, config.dst.owner_user.name
    )

    commands = []
    for c in commands_raw:
        if "NOT VALID" in command:
            if config.tables:
                regex_matches = search(
                    r"ALTER TABLE [ONLY ]*(?P<table>[a-zA-Z0-9._]+)+\s+ADD CONSTRAINT (?P<constraint>[a-zA-Z0-9._]+)+.*",
                    c,
                )
                if not regex_matches:
                    continue
                table = regex_matches.groupdict()["table"]
                if config.tables and table in config.tables:
                    commands.append(c)
            else:
                commands.append(c)

    try:
        await makedirs(schema_dir(config.db, config.dc))
    except FileExistsError:
        pass

    async with aopen(schema_file(config.db, config.dc, ONLY_INVALID), "w") as out:
        for command in commands:
            await out.write(command)

    logger.debug("Finished dumping NOT VALID constraints from the target.")


async def remove_dst_not_valid_constraints(
    config: DbupgradeConfig, logger: Logger
) -> None:
    """
    Remove the NOT VALID constraints from the schema of the target database.
    Only use if target schema was loaded in without pgbelt.
    """
    logger.info("Looking for previously dumped NOT VALID constraints...")

    async with aopen(schema_file(config.db, config.dc, ONLY_INVALID), "r") as f:
        not_valid_constraints = await f.read()

    logger.info("Removing NOT VALID constraints...")

    queries = ""
    for c in not_valid_constraints.split(";"):
        regex_matches = search(
            r"ALTER TABLE [ONLY ]*(?P<table>[a-zA-Z0-9._]+)+\s+ADD CONSTRAINT (?P<constraint>[a-zA-Z0-9._]+)+.*",
            c,
        )
        if not regex_matches:
            continue
        table = regex_matches.groupdict()["table"]
        constraint = table = regex_matches.groupdict()["constraint"]

        if (config.tables and table in config.tables) or not config.tables:
            queries = queries + f"ALTER TABLE {table} DROP CONSTRAINT {constraint};"

    command = ["psql", config.dst.owner_dsn, "-c", f"'{queries}'"]

    await _execute_subprocess(
        command, "Finished removing NOT VALID constraints.", logger
    )


async def apply_target_constraints(config: DbupgradeConfig, logger: Logger) -> None:
    """
    Load the NOT VALID constraints that were excluded from the schema. Should be called after replication during
    downtime before allowing writes into the target.
    """
    logger.info("Loading NOT VALID constraints...")

    command = [
        "psql",
        config.dst.owner_dsn,
        "-f",
        schema_file(config.db, config.dc, ONLY_INVALID),
    ]

    await _execute_subprocess(
        command, "Finished loading NOT VALID constraints.", logger
    )



from __future__ import annotations

from os.path import join
from pgbelt.util import get_logger
from pgbelt.util.asyncfuncs import makedirs
from typing import Optional

from aiofiles import open as aopen
from aiofiles.os import remove
from pydantic import BaseModel
from pydantic import ValidationError
from pydantic import validator


def config_dir(db: str, dc: str) -> str:
    return f"configs/{dc}/{db}"


def config_file(db: str, dc: str) -> str:
    return join(config_dir(db, dc), "config.json")


def not_empty(v) -> Optional[str]:
    if v == "":
        raise ValueError
    return v


class User(BaseModel):
    """
    Represents a user in a postgres db.

    name: str The user name.
    pw: str The user's password. Only required for users pgbelt needs to log in as.
    """

    name: str
    pw: Optional[str] = None

    _not_empty = validator("name", "pw", allow_reuse=True)(not_empty)


class DbConfig(BaseModel):
    """
    Represents a postgres db instance.

    host: str The hostname of this instance.
    ip: str The ip of this instance. Instance IPs must be reachable from one another.
    db: str The dbname to operate on. If you want to migrate multiple dbs in a single instance set up a separate config.
    port: str The port to connect to.
    root_user: User A superuser. Usually the postgres user.
    owner_user: User A user who owns all the data in the public schema or who has equivalent permissions. # noqa: RST301
                     This user will end up owning all the data if this is describing the target instance.
    pglogical_user: User A user for use with pglogical. Will be created if it does not exist.
    other_users: list[User] A list of other users whose passwords we might not know.
    """

    host: str
    ip: str
    db: str
    port: str
    root_user: User
    owner_user: User
    pglogical_user: User
    other_users: Optional[list[User]] = None

    _not_empty = validator("host", "ip", "db", "port", allow_reuse=True)(not_empty)

    @validator("root_user", "owner_user", "pglogical_user")
    def has_password(cls, v) -> User:  # noqa: N805
        if not v.pw:
            raise ValueError
        return v

    @property
    def root_dsn(self) -> str:
        return f"hostaddr={self.ip} port={self.port} dbname={self.db} user={self.root_user.name} password={self.root_user.pw}"

    @property
    def owner_dsn(self) -> str:
        return f"hostaddr={self.ip} port={self.port} dbname={self.db} user={self.owner_user.name} password={self.owner_user.pw}"

    @property
    def pglogical_dsn(self) -> str:
        return f"hostaddr={self.ip} port={self.port} dbname={self.db} user={self.pglogical_user.name} password={self.pglogical_user.pw}"

    @property
    def root_uri(self) -> str:
        return f"postgresql://{self.root_user.name}:{self.root_user.pw}@{self.ip}:{self.port}/{self.db}"

    @property
    def owner_uri(self) -> str:
        return f"postgresql://{self.owner_user.name}:{self.owner_user.pw}@{self.ip}:{self.port}/{self.db}"

    @property
    def pglogical_uri(self) -> str:
        return f"postgresql://{self.pglogical_user.name}:{self.pglogical_user.pw}@{self.ip}:{self.port}/{self.db}"


class DbupgradeConfig(BaseModel):
    """
    Represents a migration to be performed.

    db: str A name used to identify this specific database pair. Used in cli commands.
    dc: str A name used to identify the environment this database pair is in. Used in cli commands.
    src: DbConfig The database we are moving data out of.
    dst: DbConfig The database we are moving data into.
    """

    db: str
    dc: str
    src: Optional[DbConfig] = None
    dst: Optional[DbConfig] = None
    tables: Optional[list[str]] = None
    sequences: Optional[list[str]] = None

    _not_empty = validator("db", "dc", allow_reuse=True)(not_empty)

    @property
    def file(self) -> str:
        return config_file(self.db, self.dc)

    @property
    def dir(self) -> str:
        return config_dir(self.db, self.dc)

    async def save(self):
        """
        Write the configuration out to disk
        """
        logger = get_logger(self.db, self.dc, "config")
        logger.debug("Caching config to disk...")

        try:
            await makedirs(self.dir)
        except FileExistsError:
            pass

        try:
            await remove(self.file)
        except FileNotFoundError:
            pass

        async with aopen(self.file, "w") as f:
            await f.write(self.json(indent=4))

        logger.info("Cached config to disk.")

    @classmethod
    async def load(cls, db: str, dc: str) -> Optional[DbupgradeConfig]:
        """
        Load the specified configuration from disk if present.
        If the existing config is invalid or does not exist return None.
        """
        logger = get_logger(db, dc, "config")
        logger.debug("Trying to load cached config...")

        try:
            async with aopen(config_file(db, dc), "r") as f:
                raw = await f.read()
        except FileNotFoundError:
            logger.info("No cached config available")
            return None

        try:
            out = cls.parse_raw(raw)
        except ValidationError:
            logger.info("Cached config was not a valid DbupgradeConfig")
            return None

        logger.info("Found cached config.")

        return out



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': '.'}

packages = \
['pgbelt', 'pgbelt.cmd', 'pgbelt.config', 'pgbelt.util']

package_data = \
{'': ['*']}

install_requires = \
['aiofiles>=0.8.0,<0.9.0',
 'asyncpg>=0.26.0,<0.27.0',
 'pydantic>=1.10.1,<1.11.0',
 'tabulate>=0.8.9,<0.9.0',
 'typer>=0.6.1,<0.7.0']

entry_points = \
{'console_scripts': ['belt = pgbelt.main:app']}

setup_kwargs = {
    'name': 'pgbelt',
    'version': '0.1.0',
    'description': 'A CLI tool used to manage Postgres data migrations from beginning to end, for a single database or a fleet, leveraging pglogical replication.',
    'long_description': '# Pgbelt\n\n<p align="center">\n    <img src="https://github.com/Autodesk/pgbelt/blob/main/pgbelt.png?raw=true" width="400">\n</p>\n\n<p align="center">\n    <a href="https://github.com/autodesk/pgbelt" target="_blank">\n        <img src="https://img.shields.io/github/last-commit/autodesk/pgbelt" alt="Latest Commit">\n    </a>\n    <img src="https://github.com/Autodesk/pgbelt/actions/workflows/ci.yml/badge.svg">\n    <a href="http://www.apache.org/licenses/LICENSE-2.0" target="_blank">\n        <img src="https://img.shields.io/github/license/Autodesk/pgbelt">\n    </a>\n</p>\n\nPgBelt is a CLI tool used to manage Postgres data migrations from beginning to end,\nfor a single database or a fleet, leveraging pglogical replication.\n\nIt was built to assist in migrating data between postgres databases with as\nlittle application downtime as possible. It works in databases running different versions\nof postgres and makes it easy to run many migrations in parallel during a single downtime.\n\n| :exclamation: This is very important                                                                                                            |\n| :---------------------------------------------------------------------------------------------------------------------------------------------- |\n| As with all Data Migration tasks, **there is a risk of data loss**. Please ensure you have backed up your data before attempting any migrations |\n\n## Installation\n\n### Install From A Clone\n\nIt is recommended to install pgbelt inside a virtual environment if installing by clone:\n\n- [pyenv](https://github.com/pyenv/pyenv)\n- [pyenv-virtualenv](https://github.com/pyenv/pyenv-virtualenv)\n\nYou must also have:\n\n- [poetry](https://github.com/python-poetry/poetry)\n- Postgres Client Tools (pg_dump, pg_restore). Mac: `brew install libpq`. Ubuntu: `sudo apt-get install postgresql-client`\n\nInstall the egg locally:\n\n    # create a python virtualenv with python 3.9.11 and activate it (any 3.9.x is ok)\n    pyenv install 3.9.11\n    pyenv virtualenv 3.9.11 pgbelt\n    pyenv activate pgbelt\n\n    # Install poetry inside your virtualenv\n    pip3 install poetry\n\n    # clone the repo\n    git clone git@github.com:Autodesk/pgbelt.git\n    cd pgbelt\n\n    # install pgbelt in your virtualenv (uses poetry)\n    make install\n\n## Quickstart with Pgbelt\n\nSee [this doc](docs/quickstart.md)!\n',
    'author': 'Varjitt Jeeva',
    'author_email': 'varjitt.jeeva@autodesk.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from asyncio import gather
from pgbelt.cmd.helpers import run_with_configs
from pgbelt.config.models import DbupgradeConfig
from pgbelt.util.logs import get_logger
from pgbelt.util.postgres import analyze_table_pkeys
from pgbelt.util.postgres import precheck_info
from typing import Awaitable

from asyncpg import create_pool
from tabulate import tabulate
from typer import echo
from typer import style


async def _print_prechecks(results: list[dict]) -> list[list]:
    summary_table = [
        [
            style("database", "yellow"),
            style("server_version", "yellow"),
            style("max_replication_slots", "yellow"),
            style("max_worker_processes", "yellow"),
            style("max_wal_senders", "yellow"),
            style("pg_stat_statements", "yellow"),
            style("pglogical", "yellow"),
            style("rds.logical_replication", "yellow"),
            style("root user ok", "yellow"),
            style("owner user ok", "yellow"),
        ]
    ]

    results.sort(key=lambda d: d["db"])

    for r in results:
        root_ok = (
            r["root"]["rolcanlogin"]
            and r["root"]["rolcreaterole"]
            and r["root"]["rolinherit"]
        ) and ("rds_superuser" in r["root"]["memberof"] or r["root"]["rolsuper"])
        owner_ok = r["owner"]["rolcanlogin"]
        pg_stat_statements = (
            "installed"
            if "pg_stat_statements" in r["shared_preload_libraries"]
            else "not installed"
        )
        pglogical = (
            "installed"
            if "pglogical" in r["shared_preload_libraries"]
            else "not installed"
        )
        summary_table.append(
            [
                style(r["db"], "green"),
                style(
                    r["server_version"],
                    "green"
                    if float(r["server_version"].rsplit(" ", 1)[0].rsplit(".", 1)[0])
                    >= 9.6
                    else "red",
                ),
                style(
                    r["max_replication_slots"],
                    "green" if int(r["max_replication_slots"]) >= 20 else "red",
                ),
                style(
                    r["max_worker_processes"],
                    "green" if int(r["max_worker_processes"]) >= 20 else "red",
                ),
                style(
                    r["max_wal_senders"],
                    "green" if int(r["max_wal_senders"]) >= 20 else "red",
                ),
                style(
                    pg_stat_statements,
                    "green" if pg_stat_statements == "installed" else "red",
                ),
                style(pglogical, "green" if pglogical == "installed" else "red"),
                style(
                    r["rds.logical_replication"],
                    "green"
                    if r["rds.logical_replication"] in ["on", "Not Applicable"]
                    else "red",
                ),
                style(root_ok, "green" if root_ok else "red"),
                style(owner_ok, "green" if owner_ok else "red"),
            ]
        )

    if len(results) != 1:
        return summary_table

    # If we ran only on one db print more detailed info
    r = results[0]
    users_table = [
        [
            style("user", "yellow"),
            style("name", "yellow"),
            style("can log in", "yellow"),
            style("can make roles", "yellow"),
            style("is superuser", "yellow"),
        ]
    ]

    root_in_superusers = (
        "rds_superuser" in r["root"]["memberof"] and r["root"]["rolinherit"]
    ) or (r["root"]["rolsuper"])

    users_table.append(
        [
            style("root", "green"),
            style(r["root_name"], "green"),
            style(
                r["root"]["rolcanlogin"], "green" if r["root"]["rolcanlogin"] else "red"
            ),
            style(
                r["root"]["rolcreaterole"],
                "green" if r["root"]["rolcreaterole"] else "red",
            ),
            style(root_in_superusers, "green" if root_in_superusers else "red"),
        ]
    )

    users_table.append(
        [
            style("owner", "green"),
            style(r["owner_name"], "green"),
            style(
                r["owner"]["rolcanlogin"],
                "green" if r["owner"]["rolcanlogin"] else "red",
            ),
            style("not required", "green"),
            style("not required", "green"),
        ]
    )

    tables_table = [
        [
            style("table name", "yellow"),
            style("can replicate", "yellow"),
            style("replication type", "yellow"),
            style("schema", "yellow"),
            style("owner", "yellow"),
        ]
    ]

    for t in r["tables"]:
        can_replicate = t["Schema"] == "public" and t["Owner"] == r["owner_name"]
        replication = (
            ("pglogical" if t["Name"] in r["pkeys"] else "dump and load")
            if can_replicate
            else "unavailable"
        )
        tables_table.append(
            [
                style(t["Name"], "green"),
                style(can_replicate, "green" if can_replicate else "red"),
                style(replication, "green" if can_replicate else "red"),
                style(t["Schema"], "green" if t["Schema"] == "public" else "red"),
                style(t["Owner"], "green" if t["Owner"] == r["owner_name"] else "red"),
            ]
        )

    sequences_table = [
        [
            style("sequence name", "yellow"),
            style("can replicate", "yellow"),
            style("schema", "yellow"),
            style("owner", "yellow"),
        ]
    ]

    for s in r["sequences"]:
        can_replicate = s["Schema"] == "public" and s["Owner"] == r["owner_name"]
        sequences_table.append(
            [
                style(s["Name"], "green"),
                style(can_replicate, "green" if can_replicate else "red"),
                style(s["Schema"], "green" if s["Schema"] == "public" else "red"),
                style(s["Owner"], "green" if s["Owner"] == r["owner_name"] else "red"),
            ]
        )

    display_string = (
        style("\nDB Configuration Summary", "yellow")
        + "\n"
        + tabulate(summary_table, headers="firstrow")
        + "\n"
        + style("\nRequired Users Summary", "yellow")
        + "\n"
        + tabulate(users_table, headers="firstrow")
        + "\n"
        + style("\nTable Compatibility Summary", "yellow")
        + "\n"
        + tabulate(tables_table, headers="firstrow")
        + "\n"
        + style("\nSequence Compatibility Summary", "yellow")
        + "\n"
        + tabulate(sequences_table, headers="firstrow")
    )

    echo(display_string)

    return summary_table


@run_with_configs(skip_dst=True, results_callback=_print_prechecks)
async def precheck(config_future: Awaitable[DbupgradeConfig]) -> dict:
    """
    Report whether your source database meets the basic requirements for pgbelt.
    Any red item in a row in the table indicates a requirement not satisfied by your db.
    This command can not check network connectivity between your source and destination!

    If a dbname is given this will also show whether the configuration of
    the root and owner users seems ok and a summary of whether each
    table and sequence in the database can be replicated.
    If a row contains any red that sequence or table can not be replicated.
    """
    conf = await config_future
    pools = await gather(
        create_pool(conf.src.root_uri, min_size=1),
        create_pool(conf.src.owner_uri, min_size=1),
    )
    root_pool, owner_pool = pools

    try:
        src_logger = get_logger(conf.db, conf.dc, "preflight.src")
        result = await precheck_info(
            root_pool, conf.src.root_user.name, conf.src.owner_user.name, src_logger
        )
        result["db"] = conf.db
        result["root_name"] = conf.src.root_user.name
        result["owner_name"] = conf.src.owner_user.name
        result["pkeys"], _, _ = await analyze_table_pkeys(owner_pool, src_logger)
        return result
    finally:
        await gather(*[p.close() for p in pools])


COMMANDS = [precheck]
